'''
Base class of export. This class should be extended by NfsExportEntry or SambaExportEntry

'''
___copyright__ = "Copyright 2017 Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."

from utility import parse_boolean
from log_util import get_logger, set_log_config
from option_util import BooleanOption

LOG = get_logger('lsa.export_base')


class ExportBase(object):
    '''
    Base class of the export, this base class could be extended by NfsExportEntry or SambaExportEntry
    '''
    def __init__(self, share_path, type, **options):
        '''
        Constructor parameter:
            share_path: access point of the exported share(NFS) for example /shares/blah, resource name for samba for example samba_share_blah
            type: export type(nfs, smb)
            options: dictionary items contain export option, for example read_only(True/False)
                     for the moment the expected key is only read_only.
        '''
        if share_path is None:
            raise ValueError('Path(access point) can not be None')
        if type is None:
            raise ValueError('Export type can not be None')
        self.path = share_path
        self.type = type
        self.read_only = BooleanOption('read_only', True)
        # by default  read_only is true and options is optional, we can change the value later
        # when we process the config file i.e lazy propagation
        if options:
            ro = options.get('read_only')
            if ro is not None and parse_boolean(ro) is False:
                self.read_only.disable()

    def get_type(self):
        '''
        Get the type of the exports
        '''
        return self.type


if __name__ == '__main__':
    set_log_config()
