#!/usr/bin/bash
# Copyright (c) 2018, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.

: '
    This script is used by python backend code to build a storage pool
    parameters :
	      $1 : raid level
	      $* : devices list
'
CUR_DIR="$(dirname "$0")"
source $CUR_DIR/globals.sh
source $CUR_DIR/common.sh
source $CUR_DIR/storage.sh
source $CUR_DIR/nfs.sh
source $CUR_DIR/samba.sh

: '
    This method might be called by other scripts which have a reference to this script.
'
function main_create_pool() {
    raid_level=$1
    shift
    devices=(${*})
    echo "${FUNCNAME[0]}: raid_level=$raid_level, devices=${devices[@]}"
    out=0
    create_storage $raid_level ${devices[@]}
    out=$((out+$?))
    if [ $raid_level -ne $RAID_LEVEL_NONE ]; then
	check_raid_device
	out=$((out+$?))
	validate_raid_device
	out=$((out+$?))
    fi
    if [ $out -ne 0 ]; then
	echo "Storage setup (raid, volume group) failed"
	return 1
    fi

    create_storage_configuration
    if [ $? -ne 0 ]
    then
	echo "configuring storage setup failed"
	return 1
    fi

    setup_nfs_exports_config
    if [ $? -ne 0 ]
    then
	echo "setup of nfs exports configuration failed"
	return 1
    fi
    setup_samba_pool_config
    if [ $? -ne 0 ]
    then
	echo "setup of smb exports configuration failed"
	return 1
    fi
    setup_samba_pool_se_config
    if [ $? -ne 0 ]
    then
	echo "Samba SELinux configuration failed"
	return 1
    fi
    return 0
};
if [ $# -lt 2 ]; then
    echo "Usage : $0 <raid level> <devices1> .. <devicesn>"
    exit 1
fi
main_create_pool $@
exit $?
