#!/bin/bash
# WARNING!!. This script is for testing and development purpose only. In certain cases
# system administrator can use this script to remove the RAID setup and NFS shares.
# If you execute this script all the data stored in RAID device will be deleted.

RAID_DEV=/dev/md0
MD_PROC=/proc/mdstat
FSTAB_FILE=/etc/fstab
MD_CONF=/etc/mdadm.conf
EXPORTFS=/usr/sbin/exportfs
VGREMOVE=/usr/sbin/vgremove
PVREMOVE=/usr/sbin/pvremove
function check_exit(){
    if [ $? -ne 0 ]; then
	echo "$1 failed."
    fi
};

function get_raid_devs(){
    active=()
    if [ -f $MD_PROC ]; then
	while IFS= read -r line;
	do
	    if [[ $line  == "md0"* ]]; then
		dev=$(cut -d':' -f1 <<<"$line")
		rest=$(cut -d':' -f2 <<<"$line")
		OIFS=$IFS
		IFS=" "
		parts=($rest)
		length=${#parts[@]}
		size=$((length-2))
		for ((c=2; c<length; c++))
		do
		    str="${parts[$c]}"
		    dv=$(cut -d'[' -f1 <<<"$str")
		    active[$aval]="/dev/$dv"
		    aval=$aval+1
		done
	    fi
	done < $MD_PROC
    fi

    IFS=$'\n' active=($(sort <<<"${active[*]}"))
    echo "${active[@]}"
};

# check if raid device exists.
if [ ! -e $RAID_DEV ]; then
    echo "No raid device found($RAID_DEV): Exiting script"
    exit 0
fi

echo "Removing all the NFS shares from exports and etab"
$VGREMOVE -f lsa
check_exit "vgremove lsa"
$PVREMOVE -f /dev/md0
check_exit "pvremove"
devs=$(get_raid_devs)
echo "Stopping raid device"
mdadm --stop $RAID_DEV
check_exit "mdadm"
echo "Deleting superblock information from all devices"
mdadm --zero-superblock $devs
check_exit "mdadm"
echo "Removing entry from $MD_CONF"
sed -i '\/dev\/md0/d' $MD_CONF
check_exit "Removing entry from $MD_CONF"
echo "Removing entry from $FSTAB_FILE"
sed -i '\/dev\/md0/d' $FSTAB_FILE
check_exit "Removing entry from $MD_CONF"
echo "Successfully removed RAID setup and NFS shares"
