#!/usr/bin/bash
# Copyright (c) 2017, 2018, Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.

: '
    This script configures SElinux rules required by the web server
    and the backend wsgi daemon
'

CUR_DIR="$(dirname "$0")"
source $CUR_DIR/globals.sh
source $CUR_DIR/common.sh

: '
    Setup selinux security rules
'
function setup_selinux() {
    out=0
    # We only configure selinux if selinux state is enforcing and permissive(not disabled)
    if [ `$GETENFORCE` != 'Disabled' ]; then
	echo "set SELinux log write permissions"
	$SEMANAGE fcontext -a -t public_content_rw_t "/opt/lsa/webapp/logs(/.*)?"
	check_exit "semanage logs"
	out=$((out+$?))
	$RESTORECON -F -R -v /opt/lsa/webapp/logs
	check_exit "restorecon logs"
	out=$((out+$?))

	echo "setup SELinux configuration"
	$SETSEBOOL -P httpd_can_network_connect on
	check_exit "setsebool httpd_can_network_connect "
	out=$((out+$?))
	$SETSEBOOL -P httpd_anon_write on
	check_exit "setsebool httpd_anon_write"
	out=$((out+$?))
	$SETSEBOOL -P httpd_sys_script_anon_write on
	check_exit "setsebool httpd_sys_script_anon_write"
	out=$((out+$?))
	echo "generate custom SELinux policy"
	$CHECKMODULE -M -m -o /opt/lsa/webapp/apache/selinux-webapp-allow-rules.mod \
	    /opt/lsa/webapp/apache/selinux-webapp-allow-rules.te
	check_exit "checkmodule selinux-webapp-allow-rules.te"
	out=$((out+$?))
	$SEMOD_PKG -o /opt/lsa/webapp/apache/selinux-webapp-allow-rules.pp -m \
	    /opt/lsa/webapp/apache/selinux-webapp-allow-rules.mod
	check_exit "semmodpkg selinux-webapp-allow-rules.mod"
	out=$((out+$?))
	echo "loading custom SElinux policy"
	$SEMODULE -i /opt/lsa/webapp/apache/selinux-webapp-allow-rules.pp
	check_exit "semodule selinux-webapp-allow-rules.pp"
	out=$((out+$?))
    fi
    if [ "$out" == "0" ]; then
	return 0
    fi
    return 1
};

: '
    SElinux configuration main entry point
'
function main_setup_selinux() {
    echo "Initiating selinux setup"
    setup_selinux
    if [ $? -eq 0 ]; then
	echo "Setup selinux was successful"
    else
	echo "selinux setup failed"
    fi
    if [ "$out" == "0" ]; then
	return 0
    fi
    return 1

};

script_name=$($BASENAME ${0#-}) #- needed if sourced no path
this_script=$($BASENAME ${BASH_SOURCE})

# If running this script standalone way, should call the main function,
# Otherwise script is used as a source
if [[ ${script_name} = ${this_script} ]] ; then
    main_setup_selinux
    if [ $? == 0 ]; then
	echo "Selinux setup was successful"
	exit 0
    fi
    echo "Selinux setup failed"
    exit 1
fi
