"""
main entry point of LSA blueprint based application
- configure the flask app
- init the login manager
- register all blueprints
"""
__copyright__ = "Copyright 2018 Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."
from flask import Flask
from security import login_manager
from .jet.views import jet
from .setup.views import setup
from .lsa.views import lsa
from .login.views import login
from .bootstrap.views import bootstrap
from .common.views import common
from .doc.views import webapp_doc, auto
import os
import datetime
from backend.common import WEBAPP_ACTIVITY_TIMEOUT


class AppConfig(object):
    DEBUG = True
    SECRET_KEY = os.urandom(12)
    LOGGER_NAME = 'lsa'


app = Flask(__name__)
auto.init_app(app)

app.config.from_object(AppConfig)
app.permanent_session_lifetime = datetime.timedelta(seconds=WEBAPP_ACTIVITY_TIMEOUT)

login_manager.init_app(app)
app.register_blueprint(bootstrap, url_prefix='/')
app.register_blueprint(common, url_prefix='/common')
app.register_blueprint(jet, url_prefix='/jet')
app.register_blueprint(login, url_prefix='/login')
app.register_blueprint(lsa, url_prefix='/lsa')
app.register_blueprint(setup, url_prefix='/setup')
if os.path.isfile('/var/www/showdoc'):
    app.register_blueprint(webapp_doc, url_prefix='/doc')
