""" boostrap blueprint views module """
__copyright__ = "Copyright 2018 Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."

from flask import jsonify, Blueprint, redirect, current_app, url_for

from flask.ext.login import current_user
from ..security import is_admin_created

import backend_test as web_backend

bootstrap = Blueprint('bootstrap', __name__)

"""
The flag is checked everytime we enter here
cache the value as we do not have to perform the check each time
"""
_platform_ready = False


@bootstrap.route('/')
def index():
    """
    index URL of the LSA webapp. all requests to https://<LSA hosts> arrive here
    """
    global _platform_ready

    current_app.logger.debug('inside index of bootstrap')

    if not current_user.is_authenticated:
        current_app.logger.debug('current user not authenticated')
        if is_admin_created():
            current_app.logger.debug('redirecting to login [%s]' % url_for('login.loginuser'))
            return redirect(url_for('login.loginuser'))
        else:
            current_app.logger.debug('redirecting to [%s]' % url_for('setup.create_admin'))
            return redirect(url_for('setup.create_admin'))
    else:
        current_app.logger.debug('current user authenticated')
        if _platform_ready is False:
            # first time we enter here since server started
            p_state = web_backend.get_appliance_configuration_state()
            current_app.logger.debug('platform state [%s]' % p_state)
            _platform_ready = True if p_state == 'CONFIGURED' else False

        if _platform_ready is True:
            current_app.logger.debug('platform ready, redirecting to [%s]' % url_for('lsa.index'))
            return redirect(url_for('lsa.index'))
        else:
            # check that we are on a supported platform to allow custom
            # storage configuration
            shape = None
            try:
                shape = web_backend.get_oci_shape()
            except Exception, e:
                current_app.logger.error('cannot get current shape: %s' % str(e))
                resp = jsonify({'msg': 'cannot get current shape: %s' % str(e)})
                resp.status_code = 500
                return resp
            current_app.logger.debug('Current shape is [%s]' % shape)
            if shape.startswith('BM.Standard') or shape.startswith('VM.Standard') or shape.startswith('BM.GPU'):
                current_app.logger.debug('platform not ready, redirecting to setup')
                return redirect(url_for('setup.setup_storage_array'))
            else:
                current_app.logger.warning(
                    'not supported shape [%s] for custom pool creation' % shape)
                resp = jsonify({'msg': 'not supported shape [%s] for custom pool creation' % shape})
                resp.status_code = 500
                return resp
