/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define(['ojs/ojcore',
        'knockout',
        'jquery',
        'models/platform',
        'models/storagepool',
        'viewModels/w_vm_help-dialog',        'ojs/ojrouter',
        'ojs/ojknockout',
        'ojs/ojarraytabledatasource',
        'ojs/ojbutton',
        'ojs/ojmenu',
        'ojs/ojtoolbar',
        'ojs/ojnavigationlist',
        'ojs/ojmodule', 'ojs/ojoffcanvas'],
       function(oj, ko, $,Platform, storagePoolSVC, helpDialogVM) {
           function ControllerViewModel() {
               var self = this;

               // Media queries for repsonsive layouts
               var smQuery = oj.ResponsiveUtils.getFrameworkQuery(oj.ResponsiveUtils.FRAMEWORK_QUERY_KEY.SM_ONLY);
               self.smScreen = oj.ResponsiveKnockoutUtils.createMediaQueryObservable(smQuery);

               // Fix header on phone sizes and adjust content top margin when screen size changes.
               self.smScreen.subscribe(function() {
                   adjustContentPadding();
               });
               // Method for adjusting the content area top/bottom paddings to avoid overlap with any fixed regions.
               // This method should be called whenever your fixed region height may change.  The application
               // can also adjust content paddings with css classes if the fixed region height is not changing between
               // views.
               function adjustContentPadding() {
                   var topElem = document.getElementsByClassName('oj-applayout-fixed-top')[0];
                   var contentElem = document.getElementsByClassName('oj-web-applayout-content')[0];
                   var bottomElem = document.getElementsByClassName('oj-applayout-fixed-bottom')[0];

                   if (topElem) {
                       contentElem.style.paddingTop = topElem.offsetHeight+'px';
                   }
                   if (bottomElem) {
                       contentElem.style.paddingBottom = bottomElem.offsetHeight+'px';
                   }
               }
               // Router setup
               self.router = oj.Router.rootInstance;
               self.router.configure({
                   'dashboard': {label: 'Dashboard', isDefault: true},
                   'storage':{label:'Storage'},
                   'appliance': {label: 'Appliance'},
                   'admin': {label: 'Administration'}
               });
               oj.Router.defaults['urlAdapter'] = new oj.Router.urlParamAdapter();


               // Navigation setup
               var navData = [
                   {name: 'Dashboard', id: 'dashboard',
                    iconClass: 'oj-navigationlist-item-icon demo-icon-font-24 demo-chart-icon-24'},
                   {name: 'Storage', id: 'storage',
                    iconClass: 'oj-navigationlist-item-icon demo-icon-font-24 demo-gear-icon-16'},
                   {name: 'Appliance', id: 'appliance',
                    iconClass: 'oj-navigationlist-item-icon demo-icon-font-24 demo-gear-icon-16'},
                   {name: 'Administration', id: 'admin',
                    iconClass: 'oj-navigationlist-item-icon demo-icon-font-24 demo-copy-icon-24'}
               ];
               self.navDataSource = new oj.ArrayTableDataSource(navData, {idAttribute: 'id'});


               self.userMenuItemSelect = function( event ) {
                   if (event.target.value === 'logout') {
                       var u = new Platform().logoutUser();
                       window.location.replace(u);

                   }
                   if (event.target.value === 'changepassword') {
                       window.location.href = "/setup/changepassword";
                   }
                   if (event.target.value === 'help') {
                       helpDialogVM.show(undefined);
                   }
               };


               // User Info used in Global Navigation area
               self.userLogin = ko.observable();

               //TODO: move this to service layer
               $.getJSON('/lsa/username', function (data) {
                   self.userLogin(data.name);
               });

               this.isPlatformNotReady = ko.observable(false);
               this.platformNotAvailableMessage = ko.observable('');
               this.isStoragePoolNotReady = ko.observable(false);
               this.storagePoolNotAvailableMessage = ko.observable('');
               var isPlatformSupported = false;

               // check platform status
               new Platform().getPlatformStatus().then(function () {
                   // for supported platform (OCI shape)
                   isPlatformSupported = true;
                   self.isPlatformNotReady(false);
                   self.platformNotAvailableMessage('');
               }).catch(function (error) {
                   // for unsupported platform (OCI shape)
                   isPlatformSupported = false;
                   self.isPlatformNotReady(true);
                   self.platformNotAvailableMessage(
                       'Platform not supported (' + error +').');
               });

               // wait for 1.5 seconds as getPlatformStatus() times out in 1 second
               setTimeout(checkStoragePool, 1500);
               function checkStoragePool() {
                   // check pool status only when the platform is supported
                   if (isPlatformSupported) {
                       storagePoolSVC.getPoolStatus().then(function(poolStatus) {
                           // mirroring lsa/static/js/viewModels/storage.js isStorageDisabled logic
                           if (poolStatus.status === 'active' || poolStatus.status == 'degraded') {
                                self.isStoragePoolNotReady(false);
                                self.storagePoolNotAvailableMessage('');
                           } else {
                                self.isStoragePoolNotReady(true);
                                self.storagePoolNotAvailableMessage(
                                    'Storage pool not available. Status is '+ poolStatus.status +
                                    '. Check storage pool configuration');
                           }
                       }).catch(function(error) {
                           self.isStoragePoolNotReady(true);
                           self.storagePoolNotAvailableMessage(
                               'Storage pool not available (' + error +
                                   '). Unexpected behavior of the application.');
                       });
                   }
               }

           }

           return new ControllerViewModel();
       }
      );
