/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define(['ojs/ojcore','jquery'],
       function(oj,$) {
           function Admin() {
           }

           Admin.prototype.reboot = function() {
               return(
                       new Promise(function(resolve, reject) {
                           $.ajax({
                               'url':'/lsa/admin/reboot',
                               'async': true,
                               'dataType': "text"
                           }).done(
                               function adminRebootAjaxDone (data) {
                                   resolve(data);
                               }
                           ).fail(
                               function adminRebootAjaxFail (jqxhr, text_status, error_thown) {
                                   if (jqxhr.responseText === undefined) {
                                       // we've triggered a reboot, there are great chance
                                       // that the pipe has been broken before we get a proper response
                                       // otherwise error from the command line would have been
                                       // reported
                                       // we still resolve on that case.
                                       resolve('');
                                   }
                                   reject(jqxhr.responseText);
                               }
                           );
                       })
                   );
           };

           Admin.prototype.update = function() {
               return(
                       new Promise(function(resolve, reject) {
                           $.ajax( {
                               'url':'/lsa/admin/update',
                               'async': true,
                               'dataType': "json"
                           }).done(
                               function adminUpdateAjaxDone (data) {
                                   resolve(data);
                               }
                           ).fail(
                               function adminUpdateAjaxFail (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
                   );
           };

           Admin.prototype.restartNFSServices = function() {
               return(
                       new Promise(function(resolve, reject) {
                           $.ajax( {
                               'url':'/lsa/admin/nfs/restart',
                               'async': true,
                               'dataType': "text"
                           }).done(
                               function restartNFSServicesAjaxDone (data) {
                                   resolve(data);
                               }
                           ).fail(
                               function restartNFSServicesAjaxFail (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
                   );
           };

           Admin.prototype.restartSMBServices = function() {
               return(
                       new Promise(function(resolve, reject) {
                           $.ajax( {
                               'url':'/lsa/admin/smb/restart',
                               'async': true,
                               'dataType': "text"
                           }).done(
                              function restartSMBServicesAjaxDone (data) {
                                   resolve(data);
                               }
                           ).fail(
                               function restartSMBServicesAjaxFail (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
                   );
           };


           Admin.prototype.updateKspliceSetting =  function(cur_state) {
               return(
                       new Promise(function(resolve,reject) {
                           $.post( "/lsa/admin/ksplice/setting",
                                   {
                                       'cur_state' : cur_state
                                   },
                                   function( data ) {
                                       resolve(data);
                                   }).fail(function(jqxhr, text_status, error_thown) {
                                       reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                                    }
                                );
                        })
                    );

           };

           Admin.prototype.getSystemConfiguration = function() {
               var url = "/lsa/admin/ksplice/state";
               return(
                   new Promise(function(resolve, reject) {
                       $.ajax( {
                           'url':url,
                           'async': true,
                           'dataType': "json"
                       }).done(
                           function getKspliceEnableStateAjaxDone (data) {
                               resolve(data);
                           }
                       ).fail(
                           function getKspliceEnableStateAjaxFail (jqxhr, text_status,
                                                           error_thown) {
                               reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                           }
                       );
                   })
                );
           };

           Admin.prototype.updateSshKeys = function(key) {
              return (
                       new Promise(function(resolve,reject) {
                           $.post( "/lsa/admin/ssh/keys",
                                   {
                                       'sshkey' : key
                                   },
                                   function updateSshKeysAjaxDone( data ) {
                                       resolve(data);
                                   }).fail(function updateSshKeysAjaxFail(jqxhr, text_status, error_thown) {
                                       reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                                   });
                       }));
           };
           Admin.prototype.getSshKeys = function () {
               var url = "/lsa/admin/ssh/keys";
               return(
                   new Promise(function(resolve, reject) {
                       $.ajax( {
                           'url':url,
                           'async': true,
                           'dataType': "json"
                       }).done(
                           function getSshKeysAjaxDone (data) {
                               resolve(data.ssh_keys);
                           }
                       ).fail(
                           function getSshKeysAjaxFail (jqxhr, text_status,
                                                           error_thown) {
                               reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                           }
                       );
                   })
               );
           };

           Admin.prototype.setSambaOptions = function(options) {
              return (
                       new Promise(function(resolve,reject) {
                           $.post( "/lsa/admin/samba/options",
                                   {
                                        'options' : options
                                   },
                                   function setSambaOptionsAjaxDone( data ) {
                                       resolve(data);
                                   }).fail(function setSambaOptionsAjaxFail(jqxhr, text_status, error_thown) {
                                       reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                                   });
                       }));
           };
           Admin.prototype.getSambaOptions = function () {
               var url = "/lsa/admin/samba/options";
               return(
                   new Promise(function(resolve, reject) {
                       $.ajax( {
                           'url':url,
                           'async': true,
                           'dataType': "json"
                       }).done(
                           function getSambaOptionAjaxDone (data) {
                               resolve(data);
                           }
                       ).fail(
                           function getSambaOptionsAjaxFail (jqxhr, text_status,
                                                           error_thown) {
                               reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                           }
                       );
                   })
               );
           };

           Admin.prototype.setOCIconfiguration = function (form_data) {
               return (
                   new Promise(function(resolve, reject) {
                       $.ajax({
                           type: 'POST',
                           url: '/lsa/admin/oci/config',
                           data: form_data,
                           contentType: false,
                           cache: false,
                           processData: false,
                           async: true
                       }).done(function setOCIconfigurationAjaxDone(data) {
                           resolve(data);
                       }).fail(function setOCIconfigurationAjaxFail(jqxhr, text_status,
                                                                    error_thown) {
                           reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                       });
                   }));
           };
            Admin.prototype.getOCIconfiguration = function () {
                return(
                    new Promise(function(resolve, reject) {
                        $.ajax( {
                            'url':'/lsa/admin/oci/config',
                            'async': true,
                            'dataType': "json"
                        }).done(
                            function getOCIconfigurationAjaxDone (data) {
                                resolve(data);
                            }
                        ).fail(
                            function getOCIconfigurationAjaxFail (jqxhr, text_status,
                                                             error_thown) {
                                reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                            }
                        );
                    })
                );
            };
           Admin.prototype.clearOCIconfiguration = function() {
               return(
                       new Promise(function(resolve, reject) {
                           $.ajax( {
                               'method': 'DELETE',
                               'url':'/lsa/admin/oci/config',
                               'async': true
                           }).done(
                              function clearOCIconfigurationAjaxDone (data) {
                                   resolve(data);
                               }
                           ).fail(
                               function clearOCIconfigurationAjaxFail (jqxhr, text_status, error_thown) {
                                   reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                               }
                           );
                       })
                   );
           };
           return Admin;

       });
