/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define(['ojs/ojcore','jquery', 'models/storagepool'],
       function(oj,$, storagePoolSVC) {
           function BackupService() {

              var self = this;
               /**
                * backup a share
                * @param shareName: string representing the name of the share
                * @param namespaceName: name of the namespace of the remote object storage
                * @param bucketName: name of the bucket of the remote object storage
                * @param backupName: string backup ObjectName
                * @returns: resolve on new created task ID
                */
               this.backupShareToObjectStore  = function (shareName, namespaceName, bucketName, backupName) {
                   return(
                       new Promise(function(resolve, reject) {
                           $.post( '/lsa/shares/'+shareName+'/backup',
                                   {
                                       'namespace-name': namespaceName,
                                       'bucket-name': bucketName,
                                       'object-name': backupName
                                   },
                                   function backupShareToObjectStoreAjaxDone( data ) {
                                       storagePoolSVC.shareBackupRestoreEventchannel.dispatch(data);
                                       resolve(data);
                                   }).fail(
                                       function backupShareToObjectStoreAjaxFail (jqxhr, text_status,
                                                                                  error_thown) {
                                           reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                                       }
                                   );
                       })
                   );
               };
               /**
                * restore a share
                * @param shareName: string representing the name of the share
                * @param namespaceName: name of the namespace of the remote object storage
                * @param bucketName: name of the bucket of the remote object storage
                * @param backupName: string backup ObjectName
                */
               this.restoreShareFromObjectStore  = function (shareName, namespaceName, bucketName, backupName) {
                   return(
                       new Promise(function(resolve, reject) {
                           $.post( '/lsa/shares/'+shareName+'/restore',
                                   {
                                       'namespace-name': namespaceName,
                                       'bucket-name': bucketName,
                                       'object-name': backupName
                                   },
                                   function restoreShareToObjectStoreAjaxDone( data ) {
                                       storagePoolSVC.shareBackupRestoreEventchannel.dispatch(data);
                                       resolve(data.archive);
                                   }).fail(
                                       function restoreShareToObjectStoreAjaxFail (jqxhr, text_status,
                                                                                   error_thown) {
                                           reject(jqxhr.responseJSON ? jqxhr.responseJSON.msg : jqxhr.responseText);
                                       });
                       })
                   );
               };

           }
           return new BackupService();
       }
      );
