/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define(['ojs/ojcore',
        'knockout',
        'jquery',
        'viewModels/common_vm_def',
        'models/administration',
        'viewModels/confirmation-dialog',
        'ojs/ojinputtext',
        'ojs/ojdialog'],
       function (oj, ko, $, CommonVMDefinitons, adminSVC, confirmation)
    {
        function AddSshKeyViewModel()
        {
            var self = this;
            this.sshkeyHelp = ko.observable();


            this.sshkey = ko.observable('');
            this.statusMessage = ko.observable('');
            /**
             * method called to open ssh key addition dialog
             * @public
             */
            this.open = function ()
            {
                new adminSVC().getSshKeys().then(function(keys) {
                    self.sshkey(keys);
                    $('#txtsshkey').ojTextArea("reset");
                    $(self.dialogSelector).ojDialog('open');
                });
            };

            this.handleCancelClick = function ()
            {
                $(self.dialogSelector).ojDialog("close");
            };
            this.handleCloseDrawerClick = function ()
            {
                $(self.dialogSelector).ojDialog("close");
            };

            /**
             * Add the ssh key to the server
             * current dialog object actionType
             */
            this.handleSshKeyClick = function (data, event)
            {
                confirmation.showDialog({
                       'title':self.updateDialogTitle,
                       'message':self.updateDialogText,
                       'onOK' : function () {
                            new adminSVC().updateSshKeys(
                                self.sshkey()).then (function (payload)
                                          {
                                              oj.Logger.info('SSH key was updated successfully');
                                              confirmation.showDialog({
                                                'notifyOnly':true,
                                                'title': oj.Translations.getTranslatedString('Successful Modification'),
                                                'message': oj.Translations.getTranslatedString('Public Keys were updated successfully')
                                            });

                                          }).catch(function (why) {
                                              oj.Logger.error('Updating SSH keys failed: ' + why);
                                              confirmation.showDialog({
                                                  'notifyOnly':true,
                                                  'title': oj.Translations.getTranslatedString('SSH key modification failure'),
                                                  'message': why
                                              });
                                            }
                                          );
                            $(self.dialogSelector).ojDialog("close");
                        }
                });
            };

        }

        $.extend(AddSshKeyViewModel.prototype,CommonVMDefinitons.prototype.definitions);

        AddSshKeyViewModel.prototype.optionRequiredHint = oj.Translations.getTranslatedString('You must enter a value');


        AddSshKeyViewModel.prototype.dialogSelector='#addSshKeyDialog';
        AddSshKeyViewModel.prototype.dialogTitle=oj.Translations.getTranslatedString('Update Public SSH Keys');
        AddSshKeyViewModel.prototype.textRedLabel=oj.Translations.getTranslatedString('   SSH login is not a supported feature, this option is only provided for support and emergency recovery.')

        AddSshKeyViewModel.prototype.sshKeyHelpDef=oj.Translations.getTranslatedString('Public SSH keys of users authorized to connect to the appliance as the default opc user.');
        AddSshKeyViewModel.prototype.addKeyLabel=oj.Translations.getTranslatedString('Update keys');

        AddSshKeyViewModel.prototype.textKeyLabel=oj.Translations.getTranslatedString('Keys');

        AddSshKeyViewModel.prototype.sshKeyHelp=oj.Translations.getTranslatedString('The public SSH key will be added to authorized key file in the server');
        AddSshKeyViewModel.prototype.updateDialogTitle = oj.Translations.getTranslatedString('SSH public key update');
        AddSshKeyViewModel.prototype.updateDialogText = oj.Translations.getTranslatedString('Are you sure you want to update the SSH public keys file ?');
        return new AddSshKeyViewModel();
    }
);
