/**
 * Copyright (c) 2017, 2018, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define(['viewModels/common_vm_def', 'models/platform',
    'ojs/ojcore', 'knockout', 'jquery', 'ojs/ojknockout', 'ojs/ojgauge', ],
    function (CommonVMDefinitons, Platform, oj, ko, $)
    {


        function ApplianceViewModel()
        {
            var self = this;

            this.version = ko.observable('');
            this.versionMessage = ko.pureComputed(function ()
            {
                return self.applianceVersionLabel + ': ' + self.version();
            });

            this.applianceHostname = ko.observable('');
            this.applianceIP = ko.observable('');
            this.applianceIdentification = ko.pureComputed(function ()
            {
                return self.applianceHostname() + ' (' + self.applianceIP() + ')';
            });

            this.getlastUptimeElapsedTimeMessage = ko.pureComputed (function () {
              var howLong = Math.round(self.upTime()); // in sec
                    var howManyDays = Math.floor(howLong/86400);
                    var howManyHours = Math.floor((howLong%86400)/3600);
                    var howManyMinutes = Math.floor((howLong - (howManyDays * 86400 ) -
                                                     (howManyHours *3600 ))/60);
                    var res = '';
                    if (howManyDays > 0) {
                        res = howManyDays + ' Day' + (howManyDays > 1 ? 's' : '') + ' ';
                    }
                    if (howManyHours > 0) {
                        res = res + howManyHours + ' Hour' + (howManyHours > 1 ? 's' : '') + ' ';
                    }
                    if (howManyMinutes > 0) {
                        res = res + howManyMinutes + ' Minute' + (howManyMinutes > 2 ? 's' :'');
                    }
                    return res;
            });

            this.upTime = ko.observable();
            this.osVersion = ko.observable();
            this.kernelVersion = ko.observable();
            this.CPUCount = ko.observable();
            this.memoryCount = ko.observable();
            this.ociShape = ko.observable();
            this.instanceCreationTime = ko.observable();
            this.ociRegion = ko.observable();
            this.ociDomain = ko.observable();
            this.instanceName = ko.observable();

            function fetchAppliancePlatformInfo()
            {
                var p = new Platform();

                p.getInfos().then(
                    function (info)
                    {
                        self.upTime(info.uptime);
                        self.osVersion(info.os_version);
                        self.kernelVersion(info.kernel_version);
                        self.CPUCount(info.cpu);
                        self.memoryCount(info.mem);
                        self.version(info.platform_version);
                        self.applianceHostname(info.hostname);
                        self.applianceIP(info.ip);
                    }).catch(function (reason)
                {
                    // TODO : enough for now
                    oj.Logger.warn('get appliance info  has failed : %s', reason);
                });
            }

            this.cpuUsage = ko.observable(0);
            this.memoryUsage = ko.observable(0);
            this.swapUsage = ko.observable(0);
            this.rootFSUsage = ko.observable(0);
            function fetchApplianceUsages()
            {
                var p = new Platform();
                p.getAllUsages().then(function (usages)
                {
                    self.cpuUsage(usages.cpu);
                    self.memoryUsage(usages.mem);
                    self.swapUsage(usages.swap);
                }
                ).catch(function (reason)
                {
                    // TODO enough for now
                    oj.Logger.warn('get usages has failed : %s', reason);
                });

                p.getRootFSUsage().then(function (usage)
                {
                    self.rootFSUsage(usage);
                }).catch(function (reason)
                {
                    // TODO enough for now
                    oj.Logger.warn('get root FS usage has failed : %s', reason);
                });
            }
            function fetchInstanceDetails()
            {
                var p = new Platform();
                p.getInstanceDetails().then(function (data)
                {
                    self.ociShape(data.shape);
                    self.ociRegion(data.region);
                    self.ociDomain(data.domain);
                    self.instanceCreationTime(data.creation_time);
                    self.instanceName(data.instance_name);
                }
                ).catch(function (reason)
                {
                    oj.Logger.warn('get instance details has failed : %s', reason);
                });

            }


            this.services = ko.observableArray();
            this.servicesDataSource = new oj.ArrayTableDataSource(this.services, {idAttribute: "serviceName"});
            function fetchApplianceServices()
            {
                var p = new Platform();
                var svcs;
                p.getServices().then(
                    function (services)
                    {
                        svcs = services.map(
                            function (service)
                            {
                                return {'serviceName': service.name,
                                    'serviceDesc': service.desc,
                                    'serviceStatus': service.status,
                                    'serviceResult': service.result
                                };
                            });
                        self.services(svcs);
                    }).catch(function (reason)
                {
                    // TODO enough for now
                    oj.Logger.warn('get services status has failed : %s', reason);
                });

            }


            this.handleActivated = function (info)
            {
                if (self.version() === '') {
                    /*only the first time*/
                    fetchAppliancePlatformInfo();
                }
                fetchApplianceUsages();
                fetchInstanceDetails();
                fetchApplianceServices();
            };

        }

        $.extend(ApplianceViewModel.prototype, CommonVMDefinitons.prototype.definitions);
        $.extend(ApplianceViewModel.prototype,CommonVMDefinitons.prototype.converters);
        $.extend(ApplianceViewModel.prototype,CommonVMDefinitons.prototype.helpers);

        ApplianceViewModel.prototype.rootFSGaugeLabel =
            oj.Translations.getTranslatedString('Root filesystem utilization (%)');
        ApplianceViewModel.prototype.applianceVersionLabel =
            oj.Translations.getTranslatedString('Appliance version:');

        ApplianceViewModel.prototype.applianceServicesLabel =
            oj.Translations.getTranslatedString('Services');
        ApplianceViewModel.prototype.appliancePlatformInfoLabel =
            oj.Translations.getTranslatedString('Platform');

        ApplianceViewModel.prototype.applianceUptimeLabel =
            oj.Translations.getTranslatedString('Uptime:');
        ApplianceViewModel.prototype.applianceOsVersionLabel =
            oj.Translations.getTranslatedString('OS version:');
        ApplianceViewModel.prototype.applianceOCIShapeLabel =
                    oj.Translations.getTranslatedString('OCI shape:');
        ApplianceViewModel.prototype.applianceOCIInstanceCreationTimeLabel =
                    oj.Translations.getTranslatedString('Instance creation time:');
        ApplianceViewModel.prototype.applianceOCIInstanceNameLabel =
                    oj.Translations.getTranslatedString('Instance name:');

        ApplianceViewModel.prototype.applianceOCIInstanceRegionLabel =
                    oj.Translations.getTranslatedString('Region:');
        ApplianceViewModel.prototype.applianceOCIAvailabilityDomainLabel =
                    oj.Translations.getTranslatedString('Availability domain:');


        ApplianceViewModel.prototype.applianceKernelVersionLabel =
            oj.Translations.getTranslatedString('Kernel version:');
        ApplianceViewModel.prototype.applianceCPULabel =
            oj.Translations.getTranslatedString('CPU:');
        ApplianceViewModel.prototype.applianceMemoryLabel =
            oj.Translations.getTranslatedString('Memory:');

        ApplianceViewModel.prototype.getServiceStatusIconClasses = function (status, result)
        {
            oj.Logger.log('inside getServiceStatusIconClasses for '+status);
            var classes = ['oj-fwk-icon'];
            switch (status) {
            case 'running':
                classes.push(ApplianceViewModel.prototype.statusEnabledIcon);
                break;
            case 'exited':
                if (result == 'success') {
                    classes.push(ApplianceViewModel.prototype.statusEnabledIcon);
                } else {
                    classes.push(ApplianceViewModel.prototype.statusDisabledIcon);
                }
                break;
            case 'failed':
                classes.push(ApplianceViewModel.prototype.statusDegradedIcon);
                break;
            default:
                classes.push(ApplianceViewModel.prototype.statusUnknownIcon);
            }

            return classes.join(' ');
        };
        return new ApplianceViewModel();
    }
);
