/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define(['ojs/ojcore', 
        'ojs/ojbutton',
        'ojs/ojdialog',
        'ojs/ojpopup',
        'knockout',
        'jquery',
        'viewModels/common_vm_def',
        'ojs/ojprogress'
       ],
       function (oj, ojb, ojd, ojp, ko, $,CommonVMDefinitons)
       {
           'use strict';

           var container_class = 'portalNotifyContainerMain';

           function ConfirmationDialog()
           {

               var self = this;

               this.confirmationMessage = ko.observable('');
               
               this.confirmationTitle = ko.observable('');
               this.confirmationOkCallback = undefined;
               
               
               this.confirmationCancelCallback = undefined;
               this.confirmationNotifyOnly = ko.observable(false);
               this.confirmationHideFooter = ko.observable(false);
               this.confirmationShowProgress = ko.observable(false);
               this.confirmationLongMsg = ko.observable(false);


               /**
                * A simple modal dialog with OK and Cancel buttons.
                * @param {Object} input_opts
                * @param {string} input_opts.title - dialog title
                * @param {string} input_opts.message - dialog primary message
                * @param {string} input_opts.longMsg - dialog secondary message
                * @param {Function} [input_opts.onOK] - dialog ok callback
                * @param {Function} [input_opts.onCancel] - dialog cancel callback
                * @param {boolean} [input_opts.notifyOnly=false] -
                *  cancel button is hidden.
                * @param {boolean} [input_opts.hideFooter=false] -
                *  cancel and Ok button is hidden.
                * @param {boolean} [input_opts.showProgress=false] -
                *  show progress bar.
                * @public
                */
               this.showDialog = function (input_opts)
               {
                   $('.confirm-dialog').css('height','');
                   $('.confirm-dialog').css('width','');

                   /* Merge input_opts with default values */
                   var opts = $.extend(
                    {
                        longMsg: false,
                        notifyOnly: false,
                        hideFooter: false,
                        showProgress: false
                    },
                    input_opts);

                   this.confirmationMessage(opts.message);
                   this.confirmationTitle(opts.title);
                   this.confirmationNotifyOnly(opts.notifyOnly);
                   this.confirmationHideFooter(opts.hideFooter);
                   this.confirmationShowProgress(opts.showProgress);
                   this.confirmationLongMsg(opts.longMsg);


                   this.confirmationOkCallback = function ()
                   {
                       $('#confirmationDialog').ojDialog('close');
                       if (opts.onOK)
                       {
                           opts.onOK();
                       }
                   };
                   this.confirmationCancelCallback = function ()
                   {
                       $('#confirmationDialog').ojDialog('close');
                       if (opts.onCancel)
                       {
                           opts.onCancel();
                       }
                   };
                   $('#confirmationMSG').text(this.confirmationMessage());
                   $('#confirmationDialog').ojDialog('open');
               };
               this.hideDialog = function () {
                   $('#confirmationDialog').ojDialog('close');
               };
           }
           
           $.extend(ConfirmationDialog.prototype,CommonVMDefinitons.prototype.definitions);

           return (new ConfirmationDialog());
       }
      );
