/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define(['ojs/ojcore',
        'knockout',
        'jquery',
        'models/storagepool',
        'models/administration',
        'viewModels/common_vm_def',
        'viewModels/confirmation-dialog',
        'ojs/ojinputtext',
        'ojs/ojtable',
        'ojs/ojswitch',
        'ojs/ojdialog',
        'ojs/ojselectcombobox',
        'ojs/ojcollapsible',
        'ojs/ojknockout-validation'],
       function (oj, ko, $, storagePoolSVC, adminSVC, CommonVMDefinitons, confirmation)
    {

        function OCIConfigViewModel()
        {
            var self = this;

            this.user =  ko.observable('');
            this.tenancy =  ko.observable('');
            this.region =  ko.observable('');
            this.shouldDisableClearButton = ko.observable(false);

            this.passphrase =  ko.observable('');

            this.validationTracker = ko.observable();
            this.shouldDisableConfigureButton = ko.pureComputed(function () {
                var trackerObj = ko.utils.unwrapObservable(self.validationTracker);
                return trackerObj ? (trackerObj.invalidShown || trackerObj.invalidHidden) : false;
            });

            this.open = function ()
            {
                new adminSVC().getOCIconfiguration().then(function (configuration) {
                    self.user(configuration.user);
                    self.tenancy(configuration.tenancy);
                    self.region(configuration.region);
                    self.shouldDisableClearButton(!configuration.user || configuration.user.length === 0);
                }).catch(function(raeson) {
                    oj.Logger.warn('get OCI configuration has failed : %s',raeson);
                });
                $(self.dialogSelector).ojDialog('open');
            };

            this.handleClearClick = function (data, event)
            {
                confirmation.showDialog({
                    'title':self.clearOCIconfigDialogTitle,
                    'message':self.clearOCIconfigDialogText,
                    'onOK' : function () {
                         new adminSVC().clearOCIconfiguration();
                    }
                });
                $(self.dialogSelector).ojDialog("close");
            };
            this.handleActionClick = function (data, event)
            {
                var form_data = new FormData($('#upload-file')[0]);
                form_data.append('user', self.user());
                form_data.append('tenancy', self.tenancy());
                form_data.append('region', self.region());
                form_data.append('passphrase', self.passphrase());

                new adminSVC().setOCIconfiguration(form_data).then(function(data) {
                    oj.Logger.info('OCI service configuration updated succefully');
                    $(self.dialogSelector).ojDialog('close');
                }, function(err) {
                    confirmation.showDialog({
                        'notifyOnly':true,
                        'title': self.setConfFailed,
                        'message': err});
                });
            };
            this.handleCancelClick = function ()
            {
                $(self.dialogSelector).ojDialog("close");
            };
        }

        $.extend(OCIConfigViewModel.prototype,CommonVMDefinitons.prototype.definitions);

        OCIConfigViewModel.prototype.setConfFailed='Configuration modification has failed';

        OCIConfigViewModel.prototype.dialogSelector='#ociConfigDialog';
        OCIConfigViewModel.prototype.dialogTitle=oj.Translations.getTranslatedString('OCI service access configuration');
        OCIConfigViewModel.prototype.actionLabel=oj.Translations.getTranslatedString('Configure');
        OCIConfigViewModel.prototype.clearLabel=oj.Translations.getTranslatedString('Clear');

        OCIConfigViewModel.prototype.userOCIDLabel = oj.Translations.getTranslatedString('User OCID');
        OCIConfigViewModel.prototype.userOCIDHelpDef = oj.Translations.getTranslatedString('OCID of the user calling the API. Click for more documentation.');
        OCIConfigViewModel.prototype.userOCIDPlaceholder = oj.Translations.getTranslatedString('e.g. : ocid1.user.oc1..aaaaadcbkqyfiufvjowkl5unx4bwsoy64zdupagxefrkuhsl5lvr4g6dqipa');


        OCIConfigViewModel.prototype.tenancyLabel = oj.Translations.getTranslatedString('Tenancy OCID');
        OCIConfigViewModel.prototype.tenancyHelpDef = oj.Translations.getTranslatedString('OCID of your tenancy. Click for more documentation.');
        OCIConfigViewModel.prototype.tenancyPlaceholder = oj.Translations.getTranslatedString('e.g. : ocid1.tenancy.oc1..aaaaaaaaywf4tbgderlmotajiulowmceuzv572qbashrx2lwndezxoxrw5jq');

        OCIConfigViewModel.prototype.regionLabel = oj.Translations.getTranslatedString('Region name');
        OCIConfigViewModel.prototype.regionHelpDef = oj.Translations.getTranslatedString('An Oracle Cloud Infrastructure region. Click for more documentation.');
        OCIConfigViewModel.prototype.regionPlaceholder = oj.Translations.getTranslatedString('e.g.: us-phoenix-1');

        OCIConfigViewModel.prototype.passphraseLabel = oj.Translations.getTranslatedString('Passphrase');
        OCIConfigViewModel.prototype.passphraseHelpDef = oj.Translations.getTranslatedString('Passphrase used for the key, if it is encrypted. Click for more documentation.');


        OCIConfigViewModel.prototype.keyfileLabel = oj.Translations.getTranslatedString('Private key file');
        OCIConfigViewModel.prototype.passphraseHelpDef = oj.Translations.getTranslatedString('Private key passphrase if any. Click for more documentation.');

        OCIConfigViewModel.prototype.OCIHelpSource = 'https://docs.us-phoenix-1.oraclecloud.com/Content/API/Concepts/sdkconfig.htm';

        OCIConfigViewModel.prototype.optionRequiredHint = oj.Translations.getTranslatedString('You must enter a value');
        OCIConfigViewModel.prototype.clearOCIconfigDialogTitle = oj.Translations.getTranslatedString('clear OCI service access configuration');
        OCIConfigViewModel.prototype.clearOCIconfigDialogText = oj.Translations.getTranslatedString('Are you sure you want to clear OCI service access configuration?');

        return new OCIConfigViewModel();
    }
);
