/**
 * Copyright (c) 2017, Oracle and/or its affiliates.
 * Licensed under the Universal Permissive License v 1.0 as shown at http://oss.oracle.com/licenses/upl.
 */
define(['ojs/ojcore',
        'knockout',
        'jquery',
        'viewModels/common_vm_def',
        'models/administration',
        'viewModels/confirmation-dialog',
        'ojs/ojinputtext',
        'ojs/ojlabel',
        'ojs/ojdialog',
        'ojs/ojvalidation-base'],
       function (oj, ko, $, CommonVMDefinitons, adminSVC, confirmation)
    {
        function SambaGlobalSettingViewModel()
        {
            var self = this;
            this.workgroupName = ko.observable('');
            this.netbiosName = ko.observable('');
            this.interfaces = ko.observable('');
            this.open = function ()
            {
                new adminSVC().getSambaOptions().then(function(options)
                                {
                                    self.workgroupName(options['workgroup']);
                                    self.netbiosName(options['netbios name']);
                                    self.interfaces(options['interfaces']);
                                    $('#workgroup-name-text-input').ojInputText("reset");
                                    $('#netbios-name-text-input').ojInputText("reset");
                                    $('#interfaces-text-input').ojInputText("reset");
                                    $(self.dialogSelector).ojDialog('open');
                                }).catch(function (why)
                                         {
                                             oj.Logger.error('cannot get samba config');
                                             confirmation.showDialog(
                                                 {
                                                     'notifyOnly':true,
                                                     'title': oj.Translations.getTranslatedString('Successful Modification'),
                                                     'message': oj.Translations.getTranslatedString('Fail to get Samba configuration: ' + why)
                                                 });
                                         }
                                        );
            };
            this.workgroupValidator = {
                getHint: function () {
                    return self.workgroupValidationHintMessage;
                }
            };
            this.netbiosValidator = {
                getHint: function () {
                    return self.netbiosValidationHintMessage;
                },
                errorRaised :  ko.observable(false),
                validate: function(value) {
                    this.errorRaised(false);
                    if (value != undefined && value.length > 15) {
                        this.errorRaised(true);
                        throw new oj.ValidatorError(self.netbiosValidationHintMessage);
                    }
                    return true;
                }
            };
            this.interfacesValidator = {
                getHint: function () {
                    return self.interfacesValidationHintMessage;
                }
            };
            this.shouldDisableCreateButton = ko.pureComputed(function () {
                if (self.netbiosValidator.errorRaised() === true) {
                    return true;
                }
                return false;
            });
            this.handleCancelClick = function ()
            {
                $(self.dialogSelector).ojDialog("close");
            };
            this.handleCloseDrawerClick = function ()
            {
                $(self.dialogSelector).ojDialog("close");
            };
            this.handleUpdateClick = function (data, event)
            {
                confirmation.showDialog({
                       'title':self.updateDialogTitle,
                       'message':self.updateDialogText,
                       'onOK' : function () {
                                                        var options = {};
                                                        options['workgroup'] = self.workgroupName();
                                                        options['netbios name'] = self.netbiosName();
                                                        options['interfaces'] = self.interfaces();
                            new adminSVC().setSambaOptions(JSON.stringify(options)).then (function (payload)
                                          {
                                              oj.Logger.info('Samba global setting was updated successfully');
                                              confirmation.showDialog({
                                                'notifyOnly':true,
                                                'title': oj.Translations.getTranslatedString('Successful Modification'),
                                                'message': oj.Translations.getTranslatedString('Samba global setting was updated successfully')
                                            });

                                          }).catch(function (why) {
                                              oj.Logger.error('Updating samba global setting failed: ' + why);
                                              confirmation.showDialog({
                                                  'notifyOnly':true,
                                                  'title': oj.Translations.getTranslatedString('Samba global setting modification failure'),
                                                  'message': why
                                              });
                                            }
                                          );
                            $(self.dialogSelector).ojDialog("close");
                        }
                });
            };
        }

        $.extend(SambaGlobalSettingViewModel.prototype,CommonVMDefinitons.prototype.definitions);

        SambaGlobalSettingViewModel.prototype.dialogSelector='#sambaGlobalSettingDialog';
        SambaGlobalSettingViewModel.prototype.workgroupOptionLabel=oj.Translations.getTranslatedString('Workgroup name')
        SambaGlobalSettingViewModel.prototype.workgroupHelp=oj.Translations.getTranslatedString('Specify the workgroup that the Samba server will be added to.  Click for more information.')
        SambaGlobalSettingViewModel.prototype.workgroupPlaceHolder=oj.Translations.getTranslatedString('Name of the workgroup')
        SambaGlobalSettingViewModel.prototype.workgroupValidationHintMessage=oj.Translations.getTranslatedString('Optional and can be left empty')
        SambaGlobalSettingViewModel.prototype.netbiosOptionLabel=oj.Translations.getTranslatedString('NetBIOS name')
        SambaGlobalSettingViewModel.prototype.netbiosNameHelp=oj.Translations.getTranslatedString('Name that a Samba server is known. Click for more info')
        SambaGlobalSettingViewModel.prototype.netbiosNamePlaceHolder=oj.Translations.getTranslatedString('Name of the Samba server')
        SambaGlobalSettingViewModel.prototype.netbiosValidationHintMessage=oj.Translations.getTranslatedString('NetBIOS name cannot exceed 15 characters. When left blank, the hostname is used as the NetBIOS name.')
        SambaGlobalSettingViewModel.prototype.interfacesOptionLabel=oj.Translations.getTranslatedString('Network interfaces')
        SambaGlobalSettingViewModel.prototype.interfacesHelp=oj.Translations.getTranslatedString('Comma separated list of network interface names that Samba will use for browsing. Click for more info.')
        SambaGlobalSettingViewModel.prototype.interfacesPlaceHolder=oj.Translations.getTranslatedString('Active interfaces. eg "", "lo, eth0"')
        SambaGlobalSettingViewModel.prototype.interfacesValidationHintMessage=oj.Translations.getTranslatedString('Specify the network interfaces Samba uses for browsing. It is optional and can be left empty, in this case, all interfaces are used.')
        SambaGlobalSettingViewModel.prototype.dialogTitle=oj.Translations.getTranslatedString('Samba Global Settings')
        SambaGlobalSettingViewModel.prototype.sambaGlobalUpdateLabel=oj.Translations.getTranslatedString('Update');
        SambaGlobalSettingViewModel.prototype.updateDialogTitle = oj.Translations.getTranslatedString('Samba global setting update');
        SambaGlobalSettingViewModel.prototype.updateDialogText = oj.Translations.getTranslatedString('Are you sure you want to update Samba global settings ?');
        SambaGlobalSettingViewModel.prototype.SambaSource = 'https://www.samba.org/samba/docs/man/manpages-3/smb.conf.5.html';
        return new SambaGlobalSettingViewModel();
    }
);
