"""
module hosting all security methods
Contains admin account related method
Contains hooks used with flask-login
See methods documentation for further details
"""
__copyright__ = "Copyright 2018 Oracle.  All rights reserved."
__license__ = "Licensed under the Universal Permissive License v 1.0 as shown at  http://oss.oracle.com/licenses/upl."
from flask import jsonify, request, redirect, url_for, current_app
from flask.ext.login import LoginManager, UserMixin, AnonymousUserMixin, current_user


import logging
import datetime
import time
import pwd
import priv_exec_server as priv_exec
from backend.common import WEBAPP_ABSOLUTE_TIMEOUT

login_manager = LoginManager()

login_manager.blueprint_login_views = {
    'lsa': '/login',
    'doc': '/login',
    'setup': '/login',
    'jet': '/login'
}

_logger = logging.getLogger('lsa.login_manager')

current_users = []


# class used by flask-login to retrieve user sessions
class User(UserMixin):
    def __init__(self, username, user_key):
        self.id = unicode(user_key)
        self.name = username
        self.user_key = user_key
        self.creation_time = time.mktime(datetime.datetime.now().timetuple())

    def __eq__(self, other):
        if isinstance(other, User):
            return self.id == other.id
        return NotImplemented

    def get_id(self):
        return self.id

    def _get(id):
        for u in current_users:
            if u.id == id:
                return u
        return None


class AnonymousUser(AnonymousUserMixin):
    def __init__(self):
        AnonymousUserMixin.__init__(self)
        self.id = unicode('anonymous')
        self.user_key = 'anonymous'


login_manager.anonymous_user = AnonymousUser


@login_manager.unauthorized_handler
def unauthorized():
    ''' default view called when requesting a view that requires login '''
    _logger.warning('unauthorized')
    # if is_xhr is true that means that the request is an Ajax one. I.e coming from one of
    # our service on client side. In that case we send back Json  answer
    if request.is_xhr:
        resp = jsonify({'msg': 'unauthorized'})
        resp.status_code = 401
        return resp
    return redirect(url_for('bootstrap.index'))


def unregister_user(user):
    '''remove current user from current users list'''
    current_users.remove(user)


def register_user(user):
    '''add current user to current users list'''
    current_users.append(user)


@login_manager.user_loader
def load_user(user_id):
    ''' return the User object that matches user_id '''
    for u in current_users:
        if u.id == user_id:
            now = time.mktime(datetime.datetime.now().timetuple())
            if u.creation_time + WEBAPP_ABSOLUTE_TIMEOUT <= now:
                current_app.logger.warning('session absolute timeout')
                return None
            return u
    return None


# We may enter here each time bootstrap blueprint is called
# cache the result to avoid querying the system all the time
_is_admin_created_cached = None


def is_admin_created():
    """
    Checks that admin user is created
    return True, False otherwise
    """
    global _is_admin_created_cached
    # We may have never been here or account is not created yet
    # until the flag is not 'True' we do not cache
    if _is_admin_created_cached is None or _is_admin_created_cached is False:
        try:
            pwd.getpwnam('admin')
            _is_admin_created_cached = True
        except:
            _is_admin_created_cached = False

    return _is_admin_created_cached


def create_admin_account(password):
    '''
    Creates the admin account with the password provided
    raise:
      StandardError : if creation has failed
    '''
    res = priv_exec.priv_exec('create_admin', {'password': password,
                                               'user_key': current_user.user_key})
    if res['results'] == 'failure':
        raise StandardError(res['reason'])
